import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..Kernel import Kernel, SquaredExponential, Matern52

def test_calc_r():
    "test function for calc_r function for kernels"

    k = Kernel()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    assert_allclose(k.calc_r(x, y, params), np.array([[1., 2.], [0., 1.]]))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    assert_allclose(k.calc_r(x, y, params),
                    np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]]))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), 0.])

    assert_allclose(k.calc_r(x, y, params),
                    np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                              [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]]))

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    assert_allclose(k.calc_r(x, y, params), np.array([[1., 2.], [0., 1.]]))


def test_calc_r_failures():
    "test scenarios where calc_r should raise an exception"

    k = Kernel()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_r(x, y, params)

    with pytest.raises(FloatingPointError):
        k.calc_r(y, y, np.array([800., 0.]))

def test_calc_drdtheta():
    "test calc_drdtheta function"

    k = Kernel()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 2, 2))
    deriv[0] = 0.5*np.array([[1., 4.], [0., 1.]])/r
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([dx, 0.])))/dx

    assert_allclose(k.calc_drdtheta(x, y, params), deriv)
    assert_allclose(k.calc_drdtheta(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    r = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    deriv = np.zeros((2, 2, 2))
    deriv[0] = 0.5*np.array([[1., 4.], [0., 1.]])/r
    deriv[1] = 0.5*np.array([[4., 1.], [1., 4.]])/r
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([0., dx, 0.])))/dx

    assert_allclose(k.calc_drdtheta(x, y, params), deriv)
    assert_allclose(k.calc_drdtheta(x, y, params), deriv_fd, rtol = 1.e-5)

    deriv = np.zeros((2, 2, 2))
    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), 0.])

    r = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    deriv = np.zeros((2, 2, 2))
    deriv[0] = 0.5*2.*np.array([[1., 4.], [0., 1.]])/r
    deriv[1] = 0.5*4.*np.array([[4., 1.], [1., 4.]])/r
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([0., dx, 0.])))/dx

    assert_allclose(k.calc_drdtheta(x, y, params), deriv)
    assert_allclose(k.calc_drdtheta(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 2, 2))
    deriv[0] = 0.5*np.array([[1., 4.], [0., 1.]])/r
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.calc_r(x, y, params) -
                   k.calc_r(x, y, params - np.array([dx, 0.])))/dx

    assert_allclose(k.calc_drdtheta(x, y, params), deriv)
    assert_allclose(k.calc_drdtheta(x, y, params), deriv_fd, rtol = 1.e-5)

def test_calc_drdtheta_failures():
    "test situations where calc_drdtheta should fail"

    k = Kernel()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdtheta(x, y, params)

def test_calc_d2rdtheta2():
    "test calc_d2rdtheta2 function"

    k = Kernel()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 1, 2, 2))
    deriv[0, 0] = (0.5*np.array([[1., 4.], [0., 1.]])/r -
                   0.25*np.array([[1., 4.], [0., 1.]])**2/r**3)
    deriv_fd = np.zeros((1, 1, 2, 2))
    deriv_fd[0, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0.]))[0])/dx

    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv)
    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    r = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    deriv = np.zeros((2, 2, 2, 2))
    x12 = np.array([[1., 4.], [0., 1.]])
    x22 = np.array([[4., 1.], [1., 4.]])
    deriv[0, 0] = 0.5*x12/r-0.25*x12*x12/r**3
    deriv[0, 1] = -0.25*x12*x22/r**3
    deriv[1, 0] = -0.25*x22*x12/r**3
    deriv[1, 1] = 0.5*x22/r-0.25*x22*x22/r**3
    deriv_fd = np.zeros((2, 2, 2, 2))
    deriv_fd[0, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0., 0.]))[0])/dx
    deriv_fd[0, 1] = (k.calc_drdtheta(x, y, params)[1] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0., 0.]))[1])/dx
    deriv_fd[1, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([0., dx, 0.]))[0])/dx
    deriv_fd[1, 1] = (k.calc_drdtheta(x, y, params)[1] -
                      k.calc_drdtheta(x, y, params - np.array([0., dx, 0.]))[1])/dx

    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv)
    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), 0.])

    r = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    deriv = np.zeros((2, 2, 2, 2))
    x12 = np.array([[1., 4.], [0., 1.]])
    x22 = np.array([[4., 1.], [1., 4.]])
    deriv[0, 0] = 2.*0.5*x12/r-2.*2.*0.25*x12**2/r**3
    deriv[0, 1] = -2.*4.*0.25*x12*x22/r**3
    deriv[1, 0] = -4.*2.*0.25*x22*x12/r**3
    deriv[1, 1] = 4.*0.5*x22/r-4.*4.*0.25*x22*x22/r**3
    deriv_fd = np.zeros((2, 2, 2, 2))
    deriv_fd[0, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0., 0.]))[0])/dx
    deriv_fd[0, 1] = (k.calc_drdtheta(x, y, params)[1] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0., 0.]))[1])/dx
    deriv_fd[1, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([0., dx, 0.]))[0])/dx
    deriv_fd[1, 1] = (k.calc_drdtheta(x, y, params)[1] -
                      k.calc_drdtheta(x, y, params - np.array([0., dx, 0.]))[1])/dx

    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv)
    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 1, 2, 2))
    deriv[0, 0] = (0.5*np.array([[1., 4.], [0., 1.]])/r -
                   0.25*np.array([[1., 4.], [0., 1.]])**2/r**3)
    deriv_fd = np.zeros((1, 1, 2, 2))
    deriv_fd[0, 0] = (k.calc_drdtheta(x, y, params)[0] -
                      k.calc_drdtheta(x, y, params - np.array([dx, 0.]))[0])/dx

    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv)
    assert_allclose(k.calc_d2rdtheta2(x, y, params), deriv_fd, rtol = 1.e-5)

def test_calc_d2rdtheta2_failures():
    "test situations where calc_d2rdtheta2 should fail"

    k = Kernel()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_d2rdtheta2(x, y, params)

def test_kernel_calc_drdx():
    "test the calc_drdx method of the kernel class"

    k = Kernel()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 2, 2))
    deriv[0] = -np.array([[1., 2.], [0., 1.]])/r
    deriv_fd = np.zeros((1, 2, 2))
    # need to use central differences here as derivative is discontiuous at zero
    deriv_fd[0] = (k.calc_r(x + dx, y, params) - k.calc_r(x - dx, y, params))/dx/2.

    assert_allclose(k.calc_drdx(x, y, params), deriv)
    assert_allclose(k.calc_drdx(x, y, params), deriv_fd, rtol = 1.e-5, atol = 1.e-8)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    r = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    deriv = np.zeros((2, 2, 2))
    deriv[0] = -np.array([[1., 2.], [0., 1.]])/r
    deriv[1] = np.array([[-2., 1.], [-1., 2.]])/r
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.calc_r(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.calc_r(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.calc_r(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.calc_r(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    assert_allclose(k.calc_drdx(x, y, params), deriv)
    assert_allclose(k.calc_drdx(x, y, params), deriv_fd, rtol = 1.e-5, atol = 1.e-7)

    deriv = np.zeros((2, 2, 2))
    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), 0.])

    r = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    deriv = np.zeros((2, 2, 2))
    deriv[0] = 2.*np.array([[-1., -2.], [0., -1.]])/r
    deriv[1] = 4.*np.array([[-2., 1.], [-1., 2.]])/r
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.calc_r(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.calc_r(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.calc_r(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.calc_r(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    assert_allclose(k.calc_drdx(x, y, params), deriv)
    assert_allclose(k.calc_drdx(x, y, params), deriv_fd, rtol = 1.e-5, atol = 1.e-7)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    r = np.array([[1., 2.], [1., 1.]])

    deriv = np.zeros((1, 2, 2))
    deriv[0] = -np.array([[1., 2.], [0., 1.]])/r
    deriv_fd = np.zeros((1, 2, 2))
    # need to use central differences here as derivative is discontiuous at zero
    deriv_fd[0] = (k.calc_r(x + dx, y, params) - k.calc_r(x - dx, y, params))/dx/2.

    assert_allclose(k.calc_drdx(x, y, params), deriv)
    assert_allclose(k.calc_drdx(x, y, params), deriv_fd, rtol = 1.e-5, atol = 1.e-8)

def test_kernel_calc_drdx_failures():
    "test situations where calc_drdx should fail"

    k = Kernel()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.calc_drdx(x, y, params)

def test_squared_exponential_K():
    "test squared exponential K(r) function"

    k = SquaredExponential()

    assert_allclose(k.calc_K(1.), np.exp(-0.5))

    assert_allclose(k.calc_K(np.array([[1., 2.], [3., 4.]])),
                             np.exp(-0.5*np.array([[1., 4.], [9., 16.]])))

    with pytest.raises(AssertionError):
        k.calc_K(-1.)

def test_squared_exponential_dKdr():
    "test squared exponential dK/dr function"

    k = SquaredExponential()

    dx = 1.e-6

    assert_allclose(k.calc_dKdr(1.), -np.exp(-0.5))
    assert_allclose(k.calc_dKdr(1.), (k.calc_K(1.)-k.calc_K(1.-dx))/dx, rtol = 1.e-5)

    r = np.array([[1., 2.], [3., 4.]])

    assert_allclose(k.calc_dKdr(r), -r*np.exp(-0.5*r**2))
    assert_allclose(k.calc_dKdr(r), (k.calc_K(r)-k.calc_K(r-dx))/dx, rtol = 1.e-5)

    with pytest.raises(AssertionError):
        k.calc_dKdr(-1.)

def test_squared_exponential_d2Kdr2():
    "test squared exponential d2K/dr2 function"

    k = SquaredExponential()

    dx = 1.e-6

    assert_allclose(k.calc_d2Kdr2(1.), 0.)
    assert_allclose(k.calc_d2Kdr2(1.),
                    (k.calc_dKdr(1.)-k.calc_dKdr(1.-dx))/dx, atol = 1.e-5)

    r = np.array([[1., 2.], [3., 4.]])

    assert_allclose(k.calc_d2Kdr2(r), (r**2 - 1.)*np.exp(-0.5*r**2))
    assert_allclose(k.calc_d2Kdr2(r),
                    (k.calc_dKdr(r)-k.calc_dKdr(r-dx))/dx, rtol = 1.e-5, atol = 1.e-5)

    with pytest.raises(AssertionError):
        k.calc_d2Kdr2(-1.)

def test_squared_exponential():
    "test squared exponential covariance kernel"

    k = SquaredExponential()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    assert_allclose(k.kernel_f(x, y, params),
                    np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    assert_allclose(k.kernel_f(x, y, params),
                    np.exp(-0.5*np.array([[np.sqrt(5.), np.sqrt(5.)],
                                          [1., np.sqrt(5.)]])**2))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    assert_allclose(k.kernel_f(x, y, params),
                    2.*np.exp(-0.5*np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                                             [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])**2))

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    assert_allclose(k.kernel_f(x, y, params),
                    np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2))

def test_squared_exponential_failures():
    "test scenarios where squared_exponential should raise an exception"

    k = SquaredExponential()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_f(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_f(x, y, params)


def test_squared_exponential_deriv():
    "test the computation of the gradient of the squared exponential kernel"

    k = SquaredExponential()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    deriv = np.zeros((2, 2, 2))

    deriv[-1] = np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2)
    deriv[0] = (-0.5*np.array([[1., 4.], [0., 1.]])*
                np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2))
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    deriv = np.zeros((3, 2, 2))

    deriv[-1] = np.exp(-0.5*np.array([[np.sqrt(5.), np.sqrt(5.)],
                                      [1., np.sqrt(5.)]])**2)
    deriv[0] = -0.5*np.array([[1., 4.], [0., 1.]])*deriv[-1]
    deriv[1] = -0.5*np.array([[4., 1.], [1., 4.]])*deriv[-1]
    deriv_fd = np.zeros((3, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx, 0.])))/dx
    deriv_fd[2] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., 0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    deriv = np.zeros((3, 2, 2))

    deriv[-1] = 2.*np.exp(-0.5*np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                                         [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])**2)
    deriv[0] = -0.5*np.array([[2., 8.], [0., 2.]])*deriv[-1]
    deriv[1] = -0.5*np.array([[16., 4.], [4., 16.]])*deriv[-1]
    deriv_fd = np.zeros((3, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx, 0.])))/dx
    deriv_fd[2] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., 0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    deriv = np.zeros((2, 2, 2))

    deriv[-1] = np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2)
    deriv[0] = (-0.5*np.array([[1., 4.],[0., 1.]])*
                np.exp(-0.5*np.array([[1., 2.], [0., 1.]])**2))
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

def test_squared_exponential_deriv_failures():
    "test scenarios where squared_exponential should raise an exception"

    k = SquaredExponential()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

def test_squared_exponential_hessian():
    "test the function to compute the squared exponential hessian"

    k = SquaredExponential()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    hess = np.zeros((2, 2, 2, 2))
    r2 = np.array([[1., 4.], [0., 1.]])
    hess[0, 0] = (-0.5*r2+0.25*r2**2)*np.exp(-0.5*r2)
    hess[0, 1] = -0.5*np.exp(-0.5*r2)*r2
    hess[1, 0] = -0.5*np.exp(-0.5*r2)*r2
    hess[1, 1] = np.exp(-0.5*r2)
    hess_fd = np.zeros((2, 2, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[1])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[0])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[1])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, atol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    hess = np.zeros((3, 3, 2, 2))

    r2 = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])**2
    x12 = np.array([[1., 4.],[0., 1.]])
    x22 = np.array([[4., 1.],[1., 4.]])
    hess[0, 0] = (-0.5*x12+0.25*x12**2)*np.exp(-0.5*r2)
    hess[0, 1] = 0.25*np.exp(-0.5*r2)*x12*x22
    hess[1, 0] = 0.25*np.exp(-0.5*r2)*x12*x22
    hess[1, 1] = (-0.5*x22+0.25*x22**2)*np.exp(-0.5*r2)
    hess[0, 2] = -0.5*np.exp(-0.5*r2)*x12
    hess[2, 0] = -0.5*np.exp(-0.5*r2)*x12
    hess[1, 2] = -0.5*np.exp(-0.5*r2)*x22
    hess[2, 1] = -0.5*np.exp(-0.5*r2)*x22
    hess[2, 2] = np.exp(-0.5*r2)
    hess_fd = np.zeros((3, 3, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[1])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[0])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[1])/dx
    hess_fd[0, 2] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[0])/dx
    hess_fd[2, 0] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[2])/dx
    hess_fd[2, 1] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[2])/dx
    hess_fd[1, 2] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[1])/dx
    hess_fd[2, 2] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[2])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, atol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    hess = np.zeros((3, 3, 2, 2))

    r2 = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                   [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])**2
    x12 = np.array([[1., 4.],[0., 1.]])
    x22 = np.array([[4., 1.],[1., 4.]])
    hess[0, 0] = (-0.5*x12+0.25*x12**2)*2.*2.*2.*np.exp(-0.5*r2)
    hess[0, 1] = 0.25*2.*4.*2.*np.exp(-0.5*r2)*x12*x22
    hess[1, 0] = 0.25*np.exp(-0.5*r2)*x12*x22
    hess[1, 1] = (-0.5*x22+0.25*x22**2)*np.exp(-0.5*r2)
    hess[0, 2] = -0.5*np.exp(-0.5*r2)*x12
    hess[2, 0] = -0.5*np.exp(-0.5*r2)*x12
    hess[1, 2] = -0.5*2.*4.*np.exp(-0.5*r2)*x22
    hess[2, 1] = -0.5*2.*4.*np.exp(-0.5*r2)*x22
    hess[2, 2] = 2.*np.exp(-0.5*r2)
    hess_fd = np.zeros((3, 3, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[1])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[0])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[1])/dx
    hess_fd[0, 2] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[0])/dx
    hess_fd[2, 0] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[2])/dx
    hess_fd[2, 1] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[2])/dx
    hess_fd[1, 2] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[1])/dx
    hess_fd[2, 2] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[2])/dx

    assert_allclose(k.kernel_hessian(x, y, params)[1,2], hess[1,2])
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, atol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    hess = np.zeros((2, 2, 2, 2))
    r2 = np.array([[1., 4.], [0., 1.]])
    hess[0, 0] = (-0.5*r2+0.25*r2**2)*np.exp(-0.5*r2)
    hess[0, 1] = -0.5*np.exp(-0.5*r2)*r2
    hess[1, 0] = -0.5*np.exp(-0.5*r2)*r2
    hess[1, 1] = np.exp(-0.5*r2)
    hess_fd = np.zeros((2, 2, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[1])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[0])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[1])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, atol = 1.e-5)

def test_squared_exponential_hessian_failures():
    "test situaitons where squared_exponential_hessian should fail"

    k = SquaredExponential()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

def test_squared_exponential_inputderiv():
    "test the input derivative method of squared exponential"

    k = SquaredExponential()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    deriv = np.zeros((1, 2, 2))

    r = np.array([[1., 2.], [0., 1.]])

    deriv[0] = -r*np.exp(-0.5*r**2)*np.array([[-1., -1.], [0., -1.]])
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + dx, y, params) -
                   k.kernel_f(x - dx, y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    deriv = np.zeros((2, 2, 2))

    r = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    deriv[0] = -np.exp(-0.5*r**2)*np.array([[-1., -2.], [ 0., -1.]])
    deriv[1] = -np.exp(-0.5*r**2)*np.array([[-2.,  1.], [-1.,  2.]])
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.kernel_f(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.kernel_f(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.kernel_f(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    deriv = np.zeros((2, 2, 2))

    r = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    deriv[0] = (-2.*2.*np.exp(-0.5*r**2)*
                np.array([[-1., -2.], [0., -1.]]))
    deriv[1] = (-2.*4.*np.exp(-0.5*r**2)*
                np.array([[-2., 1.], [-1., 2.]]))
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.kernel_f(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.kernel_f(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.kernel_f(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    deriv = np.zeros((1, 2, 2))

    r = np.array([[1., 2.], [0., 1.]])

    deriv[0] = -r*np.exp(-0.5*r**2)*np.array([[-1., -1.], [0., -1.]])
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + dx, y, params)-k.kernel_f(x - dx, y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

def test_squared_exponential_inputderiv_failures():
    "test situations where input derivative method should fail"

    k = SquaredExponential()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

def test_matern_5_2_K():
    "test matern 5/2 K(r) function"

    k = Matern52()

    assert_allclose(k.calc_K(1.), (1.+np.sqrt(5.)+5./3.)*np.exp(-np.sqrt(5.)))

    r = np.array([[1., 2.], [3., 4.]])

    assert_allclose(k.calc_K(r), (1.+np.sqrt(5.)*r+5./3.*r**2)*np.exp(-np.sqrt(5.)*r))

    with pytest.raises(AssertionError):
        k.calc_K(-1.)

def test_matern_5_2_dKdr():
    "test matern 5/2 dK/dr function"

    k = Matern52()

    dx = 1.e-6

    assert_allclose(k.calc_dKdr(1.), -5./3.*(1.+np.sqrt(5.))*np.exp(-np.sqrt(5.)))
    assert_allclose(k.calc_dKdr(1.), (k.calc_K(1.)-k.calc_K(1.-dx))/dx, rtol = 1.e-5)

    r = np.array([[1., 2.], [3., 4.]])

    assert_allclose(k.calc_dKdr(r), -5./3.*r*(1.+np.sqrt(5.)*r)*np.exp(-np.sqrt(5.)*r))
    assert_allclose(k.calc_dKdr(r), (k.calc_K(r)-k.calc_K(r - dx))/dx, rtol = 1.e-5)

    with pytest.raises(AssertionError):
        k.calc_dKdr(-1.)

def test_matern_5_2_d2Kdr2():
    "test squared exponential d2K/dr2 function"

    k = Matern52()

    dx = 1.e-6

    assert_allclose(k.calc_d2Kdr2(1.), 5./3.*(5.-np.sqrt(5.)-1.)*np.exp(-np.sqrt(5.)))
    assert_allclose(k.calc_d2Kdr2(1.), (k.calc_dKdr(1.)-k.calc_dKdr(1.-dx))/dx, rtol = 1.e-5)

    r = np.array([[1., 2.], [3., 4.]])

    assert_allclose(k.calc_d2Kdr2(r), 5./3.*(5.*r**2-np.sqrt(5.)*r-1.)*np.exp(-np.sqrt(5.)*r))
    assert_allclose(k.calc_d2Kdr2(r), (k.calc_dKdr(r)-k.calc_dKdr(r - dx))/dx, rtol = 1.e-5)

    with pytest.raises(AssertionError):
        k.calc_d2Kdr2(-1.)

def test_matern_5_2():
    "test matern 5/2 covariance kernel"

    k = Matern52()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    D = np.array([[1., 2.], [0., 1.]])

    assert_allclose(k.kernel_f(x, y, params),
                    (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    D = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    assert_allclose(k.kernel_f(x, y, params),
                    (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D))

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    D = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    assert_allclose(k.kernel_f(x, y, params),
                    2.*(1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D))

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    D = np.array([[1., 2.], [0., 1.]])

    assert_allclose(k.kernel_f(x, y, params),
                    (1.+np.sqrt(5.)*D + 5./3.*D**2)*np.exp(-np.sqrt(5.)*D))

def test_matern_5_2_failures():
    "test scenarios where matern_5_2 should raise an exception"

    k = Matern52()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_f(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_f(x, y, params)

def test_matern_5_2_deriv():
    "test computing the gradient of the matern 5/2 kernel"

    k = Matern52()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    deriv = np.zeros((2, 2, 2))

    D = np.array([[1., 2.], [0., 1.]])

    deriv[-1] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)
    deriv[0] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)

    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    D = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])
    D1 = np.array([[1., 2.], [0., 1.]])
    D2 = np.array([[2., 1.], [1., 2.]])

    deriv = np.zeros((3, 2, 2))

    deriv[-1] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)
    deriv[0] = -0.5*D1**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    deriv[1] = -0.5*D2**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)

    deriv_fd = np.zeros((3, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx, 0.])))/dx
    deriv_fd[2] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., 0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    D = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])
    D1 = np.array([[1., 2.], [0., 1.]])
    D2 = np.array([[2., 1.], [1., 2.]])

    deriv = np.zeros((3, 2, 2))

    deriv[-1] = 2.*(1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)
    deriv[0] = -0.5*2.*2.*D1**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    deriv[1] = -0.5*2.*4.*D2**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)

    deriv_fd = np.zeros((3, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0., 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., dx, 0.])))/dx
    deriv_fd[2] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0., 0., dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    deriv = np.zeros((2, 2, 2))

    D = np.array([[1., 2.], [0., 1.]])

    deriv[-1] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)
    deriv[0] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)

    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([dx, 0.])))/dx
    deriv_fd[1] = (k.kernel_f(x, y, params) -
                   k.kernel_f(x, y, params - np.array([0, dx])))/dx

    assert_allclose(k.kernel_deriv(x, y, params), deriv)
    assert_allclose(k.kernel_deriv(x, y, params), deriv_fd, rtol = 1.e-5)

def test_matern_5_2_deriv_failures():
    "test scenarios where matern_5_2_deriv should raise an exception"

    k = Matern52()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_deriv(x, y, params)

def test_matern_5_2_hessian():
    "test the function to compute the squared exponential hessian"

    k = Matern52()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    D = np.array([[1., 2.], [0., 1.]])

    hess = np.zeros((2, 2, 2, 2))
    hess[0, 0] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D**4-(1.+np.sqrt(5.)*D)*D**2/2.)
    hess[0, 1] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 0] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 1] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)

    hess_fd = np.zeros((2, 2, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[0])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[1])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[1])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    D = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])
    D1 = np.array([[1., 2.], [0., 1.]])
    D2 = np.array([[2., 1.], [1., 2.]])

    hess = np.zeros((3, 3, 2, 2))

    hess[0, 0] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D1**4-(1.+np.sqrt(5.)*D)*D1**2/2.)
    hess[0, 1] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D1**2*D2**2)
    hess[1, 0] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D1**2*D2**2)
    hess[1, 1] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D2**4-(1.+np.sqrt(5.)*D)*D2**2/2.)
    hess[0, 2] = -0.5*D1**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 0] = -0.5*D1**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 2] = -0.5*D2**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 1] = -0.5*D2**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 2] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)

    hess_fd = np.zeros((3, 3, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[0])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[1])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[1])/dx
    hess_fd[0, 2] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[0])/dx
    hess_fd[2, 0] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[2])/dx
    hess_fd[1, 2] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[1])/dx
    hess_fd[2, 1] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[2])/dx
    hess_fd[2, 2] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[2])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    D = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])
    D1 = np.array([[1., 2.], [0., 1.]])
    D2 = np.array([[2., 1.], [1., 2.]])

    hess = np.zeros((3, 3, 2, 2))

    hess[0, 0] = (5./3.*2.*np.exp(-np.sqrt(5.)*D)*
                  (5./4.*2.*2.*D1**4 - (1.+np.sqrt(5.)*D)*2.*D1**2/2.))
    hess[0, 1] = 5./3.*2.*np.exp(-np.sqrt(5.)*D)*(5./4.*2.*4.*D1**2*D2**2)
    hess[1, 0] = 5./3.*2.*np.exp(-np.sqrt(5.)*D)*(5./4.*2.*4.*D1**2*D2**2)
    hess[1, 1] = (5./3.*2.*np.exp(-np.sqrt(5.)*D)*(5./4.*4.*4.*D2**4 -
                  (1. + np.sqrt(5.)*D)*4.*D2**2/2.))
    hess[0, 2] = -0.5*2.*2.*D1**2*5./3.*(1. + np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 0] = -0.5*2.*2.*D1**2*5./3.*(1. + np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 2] = -0.5*2.*4.*D2**2*5./3.*(1. + np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 1] = -0.5*2.*4.*D2**2*5./3.*(1. + np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[2, 2] = 2.*(1. + np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)

    hess_fd = np.zeros((3, 3, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[0])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[1])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[1])/dx
    hess_fd[0, 2] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[0])/dx
    hess_fd[2, 0] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0., 0.]))[2])/dx
    hess_fd[1, 2] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[1])/dx
    hess_fd[2, 1] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., dx, 0.]))[2])/dx
    hess_fd[2, 2] = (k.kernel_deriv(x, y, params)[2] -
                     k.kernel_deriv(x, y, params-np.array([0., 0., dx]))[2])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    D = np.array([[1., 2.], [0., 1.]])

    hess = np.zeros((2, 2, 2, 2))
    hess[0, 0] = 5./3.*np.exp(-np.sqrt(5.)*D)*(5./4.*D**4-(1.+np.sqrt(5.)*D)*D**2/2.)
    hess[0, 1] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 0] = -0.5*D**2*5./3.*(1.+np.sqrt(5.)*D)*np.exp(-np.sqrt(5.)*D)
    hess[1, 1] = (1.+np.sqrt(5.)*D+5./3.*D**2)*np.exp(-np.sqrt(5.)*D)

    hess_fd = np.zeros((2, 2, 2, 2))
    hess_fd[0, 0] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[0])/dx
    hess_fd[0, 1] = (k.kernel_deriv(x, y, params)[0] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[0])/dx
    hess_fd[1, 0] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([dx, 0.]))[1])/dx
    hess_fd[1, 1] = (k.kernel_deriv(x, y, params)[1] -
                     k.kernel_deriv(x, y, params-np.array([0., dx]))[1])/dx

    assert_allclose(k.kernel_hessian(x, y, params), hess)
    assert_allclose(k.kernel_hessian(x, y, params), hess_fd, rtol = 1.e-5)

def test_matern_5_2_hessian_failures():
    "test situaitons where squared_exponential_hessian should fail"

    k = Matern52()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
        k.kernel_hessian(x, y, params)

def test_matern_5_2_inputderiv():
    "test input derivative method of Matern 5/2 kernel"

    k = Matern52()

    dx = 1.e-6

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0., 0.])

    deriv = np.zeros((1, 2, 2))

    r = np.array([[1., 2.], [0., 1.]])

    deriv[0] = (-5./3.*r*(1.+np.sqrt(5.)*r)*np.exp(-np.sqrt(5.)*r)
                *np.array([[-1., -1.], [0., -1.]]))
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + dx, y, params) -
                   k.kernel_f(x - dx, y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd,
                    rtol = 1.e-5, atol = 1.e-8)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([0., 0., 0.])

    deriv = np.zeros((2, 2, 2))

    r = np.array([[np.sqrt(5.), np.sqrt(5.)], [1., np.sqrt(5.)]])

    deriv[0] = (-5./3.*(1.+np.sqrt(5.)*r)*np.exp(-np.sqrt(5.)*r)*
                np.array([[-1., -2.], [0., -1.]]))
    deriv[1] = (-5./3.*(1.+np.sqrt(5.)*r)*np.exp(-np.sqrt(5.)*r)*
                np.array([[-2., 1.], [-1., 2.]]))
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.kernel_f(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.kernel_f(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.kernel_f(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([[1., 2.], [2., 3.]])
    y = np.array([[2., 4.], [3., 1.]])
    params = np.array([np.log(2.), np.log(4.), np.log(2.)])

    deriv = np.zeros((2, 2, 2))

    r = np.array([[np.sqrt(1.*2.+4.*4.), np.sqrt(4.*2.+1.*4.)],
                  [np.sqrt(1.*4.), np.sqrt(1.*2.+4.*4.)]])

    deriv[0] = (-np.exp(2.)*np.exp(2.)*5./3.*(1.+np.sqrt(5.)*r)*
                np.exp(-np.sqrt(5.)*r)*np.array([[-1., -2.], [0., -1.]]))
    deriv[1] = (-np.exp(2.)*np.exp(4.)*5./3.*(1.+np.sqrt(5.)*r)*
                np.exp(-np.sqrt(5.)*r)*np.exp(-0.5*r**2)*np.array([[-2., 1.], [-1., 2.]]))
    deriv_fd = np.zeros((2, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + np.array([[dx, 0.], [dx, 0.]]), y, params) -
                   k.kernel_f(x - np.array([[dx, 0.], [dx, 0.]]), y, params))/dx/2.
    deriv_fd[1] = (k.kernel_f(x + np.array([[0., dx], [0., dx]]), y, params) -
                   k.kernel_f(x - np.array([[0., dx], [0., dx]]), y, params))/dx/2.

    #assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd, rtol = 1.e-5)

    x = np.array([1., 2.])
    y = np.array([2., 3.])
    params = np.array([0., 0.])

    deriv = np.zeros((1, 2, 2))

    r = np.array([[1., 2.], [0., 1.]])

    deriv[0] = (-5./3.*r*(1.+np.sqrt(5.)*r)*
                np.exp(-np.sqrt(5.)*r)*np.array([[-1., -1.], [0., -1.]]))
    deriv_fd = np.zeros((1, 2, 2))
    deriv_fd[0] = (k.kernel_f(x + dx, y, params) -
                   k.kernel_f(x - dx, y, params))/dx/2.

    assert_allclose(k.kernel_inputderiv(x, y, params), deriv)
    assert_allclose(k.kernel_inputderiv(x, y, params), deriv_fd,
                    rtol = 1.e-5, atol = 1.e-8)

def test_matern_5_2_inputderiv_failures():
    "test situations where input derivative should fail"

    k = Matern52()

    x = np.array([[1.], [2.]])
    y = np.array([[2.], [3.]])
    params = np.array([0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    params = np.array([[0., 0.], [0., 0.]])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[2., 4.], [3., 2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[1.], [2.]])
    y = np.array([[[2.], [4.]], [[3.], [2.]]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[2., 4.], [3., 2.]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

    x = np.array([[[2.], [4.]], [[3.], [2.]]])
    y = np.array([[1.], [2.]])
    params = np.array([0., 0.])

    with pytest.raises(AssertionError):
       k.kernel_inputderiv(x, y, params)

def test_Kernel_str():
    "test string method of generic Kernel class"

    k = Kernel()

    assert str(k) == "Stationary Kernel"

def test_SquaredExponential_str():
    "test string method of SquaredExponential class"

    k = SquaredExponential()

    assert str(k) == "Squared Exponential Kernel"

def test_Matern52_str():
    "test string method of Matern52 class"

    k = Matern52()

    assert str(k) == "Matern 5/2 Kernel"


from .version import version as __version__

from .GaussianProcess import GaussianProcess
from .MultiOutputGP import MultiOutputGP
from .fitting import fit_GP_MAP
from .MeanFunction import MeanFunction
from .ExperimentalDesign import MonteCarloDesign, LatinHypercubeDesign
from .SequentialDesign import MICEDesign
from .HistoryMatching import HistoryMatching
from .DimensionReduction import gKDR



import numpy as np
from mogp_emulator.GaussianProcess import GaussianProcess, PredictResult

class HistoryMatching(object):
    r"""
    Class containing tools to implement history matching between the outputs of a
    gaussian process and an observation.

    Primary usage is the ``get_implausibility`` method, which calculates the
    implausibility metric (a number of standard deviations between the emulator mean
    and the observation) for a number of query points:

    .. math::
        {I_i(\bar{x_0}) = \frac{|z_i - E(f_i(\bar{x_0}))|}
        {\sqrt{Var\left[z_i - E(f_i(\bar{x_0}))\right]}}}

    The code allows a number of ways to specify variances, and all are summed to
    compute the final variance that determines the implausibility measure.
    Variances on the observation itself can be provided with the observations,
    while variances arising from the GP prediction are included in the emulator
    expectations. Additional variances can be included to account for model
    discrepancy when computing the implausibility metric, and these can be uniform
    for all query points, or vary across different query points if desired. Any
    number of additional variances can be included in the calculation by passing
    them when computing the implausibility.

    Once implausibility is computed, the class can determine query points that
    can be ruled out, as well as points that are not ruled out yet, based on the
    threshold class attribute. See the methods ``get_RO``, ``get_NROY``, and
    ``set_threshold`` for more details.

    Example - implausibility computation for a 1D GP::

        import math
        import numpy as np
        from HistoryMatching import HistoryMatching
        from mogp_emulator import GaussianProcess

        # Set up a gaussian process
        x_training = np.array([[0.],[10.],[20.],[30.],[43.],[50.]])
        def get_y_training(x):
            n_points = len(x)
            f = np.zeros(n_points)
            for i in range(n_points):
                f[i] = np.sin(2*math.pi*x[i] / 50)
            return f
        y_training = get_y_training(x_training)

        gp = GaussianProcess(x_training, y_training)
        np.random.seed(47)
        gp.learn_hyperparameters()

        # Define the observation to which to compare
        obs = [-0.8, 0.0004]

        # Coordinates to predict
        n_rand = 2000
        coords = np.sort(np.random.rand(n_rand),axis=0)*(56)-3
        coords = coords[:,None]

        # Generate Expectations
        expectations = gp.predict(coords)

        # Calculate implausibility
        hm = HistoryMatching(obs=obs, expectations=expectations)
        I = hm.get_implausibility()


    """

    def __init__(self, gp=None, obs=None, coords=None, expectations=None, threshold=3.):
        r"""
        Create a new instance of history matching.

        The primary function of this method is to initialise the ``self.gp``,
        ``.obs``, ``.coords``, ``.expectations``, ``.ndim``, ``.ncoords``,
        ``.threshold``, ``.I``, ``.NROY``, and ``.RO`` variables as placeholders.
        The optional keyword arguments can also be used to set user-defined values for
        various quantities in a single call to ``HistoryMatching()``. Setting sufficient
        numbers of these to allow the number of dimensions and/or number of coordinates to
        be computed also causes these to be set at this stage.

        :param gp: (Optional) ``GaussianProcess`` object used to make predictions in
                   history matching. Optional, can instead provide predictions directly
                   via the ``expectations`` argument. Default is ``None``
        :type gp: GaussianProcess or None
        :param obs: (Optional) Observations against which the predictions will be
                    compared. If provided, must either be a float (assumes no
                    uncertainty in the observations) or a list of two floats
                    holding the observation and its variance. Required for history
                    matching, but can be provided when calling ``get_implausibility``.
                    Default is ``None``.
        :type obs: float, list, or None
        :param coords: (Optional) Inputs at which the emulator values will be computed
                       to compare the GP output to the observations. If provided, must
                       be a numpy array matching the input for the GP (must be a 2D
                       array with dimensions ``(n, D)`` where ``n`` is the number of
                       points to be considered and ``D`` is the number of parameters
                       for the GP). Only required if ``expectations`` is not provided.
                       Default is ``None``.
        :type coords: ndarray
        :param expectations: (Optional) Tuple of 3 numpy arrays or 2 numpy arrays and
                             ``None`` of the form expected from GP predictions. The first
                             array must hold the predicted mean at all query points, and
                             the second array must hold the variances from the GP predictions
                             at all query points. The third is not used, so can simply be a
                             dummy array or ``None``. Can instead provide a GP object and the
                             points to query to have the predictions made internally. Default
                             is ``None``.
        :type expectations: tuple holding 3 ndarrays
        """

        # Place-holder values for user-defined quantities that can be passed in
        # as args.
        self.gp = None
        self.obs = None
        self.coords = None
        self.expectations = None

        # Place-holder or default values for optional or derived quantities.
        self.ndim = None
        self.ncoords = None
        self.threshold = None
        self.I = None
        self.NROY = None
        self.RO = None

        # If suitable keyword arguments are provided, update the relevent
        # variable(s)
        if self.check_gp(gp):
            self.set_gp(gp)
        if self.check_obs(obs):
            self.set_obs(obs)
        if self.check_coords(coords):
            self.set_coords(coords)
        if self.check_expectations(expectations):
            self.set_expectations(expectations)
        if self.check_threshold(threshold):
            self.set_threshold(threshold)

        self.update()


    def _select_expectations(self):
        r"""
        Determine the emulator predictions to use in HistoryMatching

        This method queries the internal state of the HistoryMatching object to get
        the expectations to use in the history matching calculation. It either
        uses the expectations that were directly provided, or makes predictions
        at the query points using the provided GP. If both methods are set
        simultaneously, the method raises an exception.

        :returns: Expectations to use in history matching. These are either provided
                  explicitly, or made internally using the GP and query points.
        :rtype: tuple of 3 ndarrays or 2 ndarrays and None
        """
        # Check that we have exactly 1 valid combination of parameters
        UseCoordGP = False
        UseExpectations = False
        if (self.check_coords(self.coords) and self.check_gp(self.gp)):
            UseCoordGP = True
        if self.check_expectations(self.expectations):
            UseExpectations = True
        if UseCoordGP and UseExpectations:
            raise ValueError("Multiple valid parameter combinations are set. Previously set " +
                             "parameters can be removed by setting them to None")
        if not UseCoordGP and not UseExpectations:
            raise ValueError("Expectations are not provided, nor is a GP and coordinates." +
                             "must set one in order to perform History Matching")

        # Confirm that the ncoords parameter is set
        if self.ncoords is None:
            raise ValueError("ncoords is not set despite a valid parameter combination "+
                             "being found.")

        # From this point on, the calculation works off the self.expectations value.
        # If this is not set, calculate it from the other parameters
        if UseCoordGP:
            expectations = self.gp.predict(self.coords)
        else:
            expectations = self.expectations

        return expectations

    def get_implausibility(self, discrepancy = 0.):
        r"""
        Compute Implausibility measure for all query points

        Carries out the implausibility calculation given by:

        .. math::
            {I_i(\bar{x_0}) = \frac{|z_i - E(f_i(\bar{x_0}))|}
            {\sqrt{Var\left[z_i - E(f_i(\bar{x_0}))\right]}}}

        to return an implausibility value (a number of standard deviations between the
        emulator mean and the observation)for each of the provided coordinates.

        Requires that the observation parameter is set, and that at least one of the
        following conditions are met:

          a) The coords and gp parameters are set
          b) The expectations parameter is set

        Note that using the GP internally assumes that the standard prediction settings
        will be used when making predictions. If the user wishes to have more control
        over the prediction method (i.e. make the predictions from MCMC samples),
        the user should explicitly pass ``expectations`` to the object.

        An additional variance can be included that represents a general model
        discrepancy that describes the prior beliefs regarding how well the model
        matches reality. In practice, the model discrepancy is essential to have
        predictions that are not overly confident, however it can be hard to estimate
        (see further discussion in Brynjarsdottir and O\'Hagan, 2014).

        As the implausibility calculation linearly sums variances, the result is
        agnostic to the precise provenance of any of the included variances

        :param discrepancy: Additional variance to be included in the implausibility
                            calculation. Must be a non-negative float. Optional, default
                            is ``0.``
        :type discrepancy: float
        :returns: Array holding implausibility metric for all query points accounting
                  for all variances, ndarray of length ``(ncoords,)``
        :rtype: ndarray
        """

        # Confirm that observation parameter is set
        if not self.check_obs(self.obs):
            raise ValueError("implausibility calculation requires that the observation value is " +
                             "set. This can be done using the set_obs method.")

        assert discrepancy >= 0., "Model discrepancy variance cannot be negative"

        expectations = self._select_expectations()

        # Compute implausibility for each expectation value
        Vs = np.zeros(self.ncoords)
        Vs += expectations[1]                         # variance on expectation
        Vs += discrepancy                             # model discrepancy
        Vs += self.obs[1]                             # variance on observation
        self.I = (np.abs(self.obs[0] - expectations[0]) / np.sqrt(Vs))

        return self.I


    def get_NROY(self, discrepancy = 0.):
        r"""
        Return set of indices that are not yet ruled out

        Returns a list of indices for ``self.I`` that correspond to entries that are not
        yet ruled out. Points that are ruled out have an implausibility metric that
        exceeds the threshold (can be set when initializing history matching or using
        the ``set_threshold`` method). If the implausibility metric has not yet been
        computed for the desired points, it is calculated. If a model discrepancy
        is to be included, it can be passed here.

        :param discrepancy: Additional variance to be included in the implausibility
                            calculation. Must be a non-negative float. Optional, default
                            is ``0.``
        :type discrepancy: float
        :returns: List of integer indices that have not yet been ruled out.
        :rtype: list
        """
        if self.I is None:
            self.get_implausibility(discrepancy)

        self.NROY = list(np.where(self.I <= self.threshold)[0])

        return self.NROY


    def get_RO(self, discrepancy = 0.):
        r"""
        Return set of indices that have been ruled out

        Returns a list of indices for ``self.I`` that correspond to entries that have
        been ruled out. Points that are ruled out have an implausibility metric that
        exceeds the threshold (can be set when initializing history matching or using
        the ``set_threshold`` method). If the implausibility metric has not yet been
        computed for the desired points, it is calculated. If a model discrepancy
        is to be included, it can be passed here.

        :param discrepancy: Additional variance to be included in the implausibility
                            calculation. Must be a non-negative float. Optional, default
                            is ``0.``
        :type discrepancy: float
        :returns: List of integer indices that have been ruled out.
        :rtype: list
        """
        if self.I is None:
            self.get_implausibility(discrepancy)

        self.RO = list(np.where(self.I > self.threshold)[0])

        return self.RO


    def set_gp(self, gp):
        r"""
        Set the Gaussian Process to use with history matching

        Sets the ``self.gp`` variable to the provided ``GaussianProcess`` argument.

        :param gp: ``GaussianProcess`` object to use for history matching.
        :type gp: GaussianProcess
        :returns: None
        """
        if not self.check_gp(gp):
            raise TypeError("bad input for set_gp - expects a GaussianProcess object.")
        self.gp = gp


    def set_obs(self, obs):
        r"""
        Set the observations to be used for history matching

        Sets the ``self.obs`` variable to the provided ``obs`` argument. The object
        must pass the ``check_obs`` requirements, meaning that it must be a
        float (assumes that the observation has no error associated with it) or a
        list containing two floats (representing an observation and its variance).
        Note that the variance must be non-negative.

        :param obs: Observations to be used for history matching. Must be a float
                    or a list of two floats.
        :type obs: float or list
        :returns: None
        """
        if not self.check_obs(obs):
            raise TypeError("bad input for set_obs")
        if isinstance(obs, list):
            if len(obs) == 1:
                self.obs = [float(obs[0]), 0.]
            else:
                self.obs = [float(a) for a in obs]
        else:
            self.obs = [float(obs), 0.]


    def set_coords(self, coords):
        r"""
        Set the query points to be used in history matching

        Sets the ``self.coords`` variable to the provided query points ``coords`` if
        it passes the ``check_coords`` requirements, or be ``None`` to remove a set
        of existing query points. ``coords`` must be a numpy array matching the
        inputs to the provided ``GaussianProcess`` object.

        :param coords: Numpy array holding query points (array with shape ``(n, D)``,
                       where ``n`` is the number of query points and ``D`` is the
                       number of inputs to the emulator), or ``None``
        :type coords: ndarray or None
        :returns: None
        """
        # need to allow coords == None, as otherwise can't reset values
        if not self.check_coords(coords) and (not coords is None):
            raise TypeError("bad input for set_coords - expected coords in the form " +
                            "of a list or 1D or 2D ndarray of numerical values")
        if isinstance(coords, np.ndarray):
            if len(coords.shape) == 1:
                self.coords = np.reshape(coords, [-1, 1])
            elif len(coords.shape) == 2:
                self.coords = coords
            else:
                raise Exception("error in exception handling - an ndarray of >2 " +
                                "dimensions somehow got through check_coords")
        elif isinstance(coords, list):
            self.coords = np.reshape(np.asarray(coords), [-1, 1])
        elif coords is None:
            self.coords = None
        else:
            raise TypeError("error in exception handling - an argument of illegal " +
                            "type somehow got through check_coords")
        self.update()


    def set_expectations(self, expectations):
        r"""
        Set the expected output of the simulator to be used in history matching

        Sets the ``self.expectations`` variable to the provided ``expectations``
        argument if it passes the ``check_expectations`` requirements, or ``None``
        can be used to remove an existing set of expectations. Expectations
        must be a tuple of 3 numpy arrays, where the first holds the predicted
        means for all query points and the second holds the predicted variances
        for all query points. The third arrray is not used in the computation,
        but is included as it is an expected output of the GP ``predict`` method.

        :param expectations: GP predictions at all query points. Must be a tuple
                             of 3 numpy arrays, or None to remove existing
                             expectations.
        :type expectations: tuple of 3 ndarrays or None
        :returns: None
        """
        # need to allow expectations == None, as otherwise can't reset values
        if not self.check_expectations(expectations) and (not expectations is None):
            raise TypeError("bad input for set_expectations - expected a Tuple " +
                            "of 3 ndarrays.")
        self.expectations = expectations
        self.update()


    def set_threshold(self, threshold):
        r"""
        Set the threshold value for history matching

        Sets the ``self.threshold variable`` to the provided threshold argument
        if it  passes the ``check_threshold`` requirements. The threshold must
        be a non-negative float.

        :param threshold: New value for threshold, must be a non-negative float.
        :type threshold: float
        :returns: None
        """
        if not self.check_threshold(threshold):
            raise TypeError("bad input for set_expectations - expected a float")
        self.threshold = float(threshold)


    def status(self):
        r"""
        Prints a summary of the current status of the class object.

        Prints a summary of the current status of the class object, including the
        values stored in ``self.gp``, ``.obs``, ``.ndim``, ``.ncoords``, and
        ``.threshold``, and the shapes/sizes ofvalues stored in ``self.obs``,
        ``.coords``, ``.expectations``, ``.I``, ``.NROY``, and ``.RO.``
        No inputs, no return value.

        :returns: None
        """

        print(str(self))


    def check_gp(self, gp):
        r"""
        Checks if the provided argument is consistent with expectations for a GP.

        Returns a boolean that is True if the provided quantity is consistent with
        the requirements for a gaussian process, i.e. is of type ``GaussianProcess``.

        :param gp: Input GP object to be checked.
        :type gp: GaussianProcess
        :returns: Boolean indicating if provided object is a ``GaussianProcess``
        :rtype: bool
        """
        if gp is None: return False
        if isinstance(gp, GaussianProcess):
            return True
        return False


    def check_obs(self, obs):
        r"""
        Checks if the provided argument is consistent with expectations for
        observations.

        Returns a boolean that is ``True`` if the provided quantity is consistent with
        the requirements for the observation quantity, i.e. is a numerical value
        that can be converted to a float, or a list of up to two such values. Also
        checks if the provided variance is non-negative.

        :param obs: Input for observations to be checked
        :type obs: float or list-like
        :returns: Boolean indicating if provided observations are acceptable
        :rtype: bool
        """

        if obs is None:
            return False
        if isinstance(obs, list) or isinstance(obs, tuple) or isinstance(obs, np.ndarray):
            if len(obs) > 2:
                raise ValueError("bad input type for HistoryMatching - the specified " +
                                 "observation parameter cannot contain more than 2 entries "+
                                 "(value, [variance])")
            if len(obs) <= 2:
                try:
                    test = [float(a) for a in obs]
                except TypeError:
                    raise TypeError("bad input type for HistoryMatching - the specified " +
                                    "observation parameter must contain only numerical " +
                                    "values")
            if len(obs) == 2:
                assert float(obs[1]) >= 0., "variance in observations cannot be negative"
            return True
        else:
            try:
                test = float(obs)
            except (TypeError, ValueError):
                raise TypeError("bad input type for HistoryMatching - the specified " +
                                "observation parameter must contain only numerical values")
            return True
        return False


    def check_coords(self, coords):
        r"""
        Checks if the provided argument is consistent with expectations for
        coordinates.

        Returns a boolean that is ``True`` if the provided quantity is consistent with
        the requirements for the coordinates quantity, i.e. a ndarray of
        fewer than 3 dimensions.

        :param coords: Input to check for consistency with coordinates.
        :type coords: ndarray
        :returns: Boolean indicating if coords is consistent
        :rtype: bool
        """

        if coords is None:
            return False

        if isinstance(coords, np.ndarray):
            if len(coords.shape) <= 2:
                return True
        return False


    def check_expectations(self, expectations):
        r"""
        Checks if the provided argument is consistent with expectations for
        Gaussian Process Expectations.

        Returns a boolean that is ``True`` if the provided quantity is consistent with
        the output of the predict method of a GaussianProcess object, i.e. that it
        is a ``GaussianProcess.PredictResult`` object with mean and variance defined.

        :param expectations: Input to check for consistency with expectations
        :type expectations: tuple of 3 numpy arrays or 2 numpy arrays and None
        :returns: Boolean indicating if expectations is consistent
        :rtype: bool
        """
        if expectations is None:
            return False
        if not isinstance(expectations, PredictResult):
            return False
        if not all((isinstance(expectations[0], np.ndarray),
                    isinstance(expectations[1], np.ndarray),
                    (isinstance(expectations[2], np.ndarray) or
                     expectations[2] is None))):
            raise TypeError("bad input type for HistoryMatching - expected expectation " +
                            "values in the form of a PredictResults object with mean" +
                            "and uncertainty set.")
        if not expectations[0].shape == expectations[1].shape:
            raise ValueError("bad input for HistoryMatching - mean and variance " +
                             "expectations do not match")
        assert np.all(expectations[1] >= 0.), "all variances must be non-negative"
        return True


    def check_threshold(self, threshold):
        r"""
        Check value of threshold

        Checks if the provided argument is consistent with expectations for a
        threshold value.

        Returns a boolean that is ``True`` if the provided quantity is consistent with
        the requirements for a threshold value, i.e. that it is a non-negative numerical
        value.

        :param threshold: threshold to be tested
        :type threshold: float
        :returns:  Boolean indicating if the provided argument is consistent with
                   a threshold
        :rtype: bool
        """
        if threshold is None:
            return False
        try:
            test = float(threshold)
            assert test >= 0., "threshold must be non-negative"
            return True
        except TypeError:
            return False

    def update(self):
        r"""
        Update History Matching object

        Checks that sufficient information exists to compute ``ndim`` and/or ``ncoords``
        and, if so, computes these values. This also serves to update these values
        if the information on which they are based is changed. No inputs, no return
        value.

        :returns: None
        """

        if self.check_coords(self.coords):
            self.ndim = self.coords.shape[1]
            self.ncoords = self.coords.shape[0]
        elif self.check_expectations(self.expectations):
            self.ncoords = self.expectations[0].shape[0]


    def __str__(self):
        r"""
        Returns string representation of ``HistoryMatching`` object

        Returns a string representation of a ``HistoryMatching`` object,
        used when printing information to the interpreter or when the
        ``status`` method is called.

        :returns: string representation of the HistoryMatching object.
        :rtype: str
        """

        if self.coords is None:
            coord_str = None
        else:
            coord_str = self.coords.shape

        if self.expectations is None:
            exp_str = None
        else:
            exp_str = (str(type(self.expectations)) + " of len " +
                       str(len(self.expectations)) + " containing arrays of shape: " +
                       str(self.expectations[0].shape))

        if self.I is None:
            I_str = None
        elif isinstance(self.I, np.ndarray):
            I_str = (str(type(self.I)) + " of shape " + str(self.I.shape))

        if self.NROY is None:
            NROY_str = None
        elif isinstance(self.NROY, list):
            NROY_str = (str(type(self.NROY)) + " of length " + len(self.NROY))

        if self.RO is None:
            RO_str = None
        elif isinstance(self.RO, list):
            RO_str = (str(type(self.RO)) + " of length " + len(self.RO))

        return ("History Matching tools created with:\n" +
                "Gaussian Process: {}\n" +
                "Observations: {}\n" +
                "Coords: {}\n" +
                "Expectations: {}\n" +
                "No. of Input Dimensions: {}\n" +
                "No. of Descrete Expectation Values: {}\n" +
                "I_threshold: {}\n" +
                "I: {}\n" +
                "NROY: {}\n" +
                "RO: {}\n").format(self.gp, self.obs, coord_str, exp_str,
                                   self.ndim, self.ncoords, self.threshold,
                                   I_str, NROY_str, RO_str)



import numpy as np
import scipy
from scipy.integrate import solve_ivp

assert scipy.__version__ >= '1.4', "projectile.py requires scipy version 1.4 or greater"

# Create our simulator, which solves a nonlinear differential equation describing projectile
# motion with drag. A projectile is launched from an initial height of 2 meters at an
# angle of 45 degrees and falls under the influence of gravity and air resistance.
# Drag is proportional to the square of the velocity. We would like to determine the distance
# travelled by the projectile as a function of the drag coefficient and the launch velocity.

# define functions needed for simulator

def f(t, y, c):
    "Compute RHS of system of differential equations, returning vector derivative"

    # check inputs and extract

    assert len(y) == 4
    assert c >= 0.

    vx = y[0]
    vy = y[1]

    # calculate derivatives

    dydt = np.zeros(4)

    dydt[0] = -c*vx*np.sqrt(vx**2 + vy**2)
    dydt[1] = -9.8 - c*vy*np.sqrt(vx**2 + vy**2)
    dydt[2] = vx
    dydt[3] = vy

    return dydt

def event(t, y, c):
    "event to trigger end of integration"

    assert len(y) == 4
    assert c >= 0.

    return y[3]

event.terminal = True

# now can define simulator

def simulator(x):
    "simulator to solve ODE system for projectile motion with drag. returns distance projectile travels"

    # unpack values

    assert len(x) == 2
    assert x[1] > 0.

    c = 10.**x[0]
    v0 = x[1]

    # set initial conditions

    y0 = np.zeros(4)

    y0[0] = v0/np.sqrt(2.)
    y0[1] = v0/np.sqrt(2.)
    y0[3] = 2.

    # run simulation

    results = solve_ivp(f, (0., 1.e8), y0, events=event, args = (c,))

    return results.y_events[0][0][2]

# function for printing out results

def print_results(inputs, predictions):
    "convenience function for printing out results and computing mean square error"

    print("Target Point                   Predicted mean            Actual Value")
    print("------------------------------------------------------------------------------")

    error = 0.

    for pp, m in zip(inputs, predictions):
        trueval = simulator(pp)
        print("{}      {}       {}".format(pp, m, simulator(pp)))
        error += (trueval - m)**2

    print("Mean squared error: {}".format(np.sqrt(error)/len(predictions)))


from mogp_emulator import MeanFunction
try:
    from patsy import ModelDesc, Term, EvalFactor
    no_patsy = False
except ImportError:
    no_patsy = True

def mean_from_patsy_formula(formula, inputdict={}):
    """
    Create a mean function from a patsy formula

    This is the functional interface to creating a mean function from a patsy/R formula.
    The higher level function ``MeanFunction`` in the ``MeanFunction`` module is preferred
    over this one, but this potentially gives the user slightly more control as a
    patsy ``ModelDesc`` object can be passed directly, rather than giving a string.

    This method takes a string or a patsy ``ModelDesc`` object as an input and an optional
    dictionary that map strings to integer indices in the input data. The formula is then
    parsed with patsy, and the individual terms resulting from that are converted to
    mean functions and composed using the provided operations.

    The string formulas can be specified in several ways. The formula LHS is implicitly
    always ``"y = "`` or ``"y ~ "``, though these can be explicitly provided as well
    (though it is ignored in the conversion). The RHS may contain a set of terms
    containing the add, multiply, power, and call operations much in the same way
    that the operations would be entered as regular python code. Parentheses are
    used to indicated prececence as well as the call operation, and square brackets
    indicate an indexing operation on the inputs. Inputs may be specified as either
    a string such as ``"x[0]"``, ``"inputs[0]"``, or a string that can be mapped to
    an integer index with the optional dictionary passed to the function. Any strings
    not representing operations or inputs as described above are interpreted as follows:
    if the string can be converted into a number, then it is interpreted as a
    ``ConstantMean`` fixed mean function object; otherwise it is assumed to represent
    a fitting coefficient. Note that this means many characters that do not represent
    operations within this mean function language but would not normally be considered
    as python variables will nonetheless be converted into fitting coefficients --
    it is up to the user to get this right.

    Expressions that are repeated or redundant will not be simplified beyond the parsing
    done by patsy, so the user should take care that the provided expression is sensible
    as a mean function and will not cause problems when fitting.

    Examples: ::

        >>> from mogp_emulator.formula import mean_from_patsy_formula
        >>> mf1 = mean_from_patsy_formula("x[0]")
        >>> print(mf1)
        c + c*x[0]
        >>> mf2 = mean_from_patsy_formula("a*b", {"a": 0, "b": 1})
        >>> print(mf2)
        c + c*x[0] + c*x[1] + c*x[0]*x[1]

    :param formula: string representing the desired mean function formula
                    or a patsy ``ModelDesc`` object
    :type formula: str or ModelDesc
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: New subclass of ``MeanBase`` implementing the given formula
    :rtype: subclass of MeanBase (exact type will depend on the formula that is provided)
    """

    assert not no_patsy, "patsy must be installed to parse formulas using patsy"

    if isinstance(formula, str):
        model = ModelDesc.from_formula(formula)
    elif isinstance(formula, ModelDesc):
        model = formula

    model_terms = []

    for term in model.rhs_termlist:
        model_terms.append(_term_to_mean(term, inputdict))

    mf = model_terms.pop(0)

    assert issubclass(type(mf), MeanFunction.MeanBase)

    for term in model_terms:
        mf += term

    assert issubclass(type(mf), MeanFunction.MeanBase)

    return mf

def mean_from_string(formula, inputdict={}):
    """
    Create a mean function from a string formula

    This is the functional interface to creating a mean function from a string formula.
    The higher level function ``MeanFunction`` in the ``MeanFunction`` module is preferred
    over this one, but this can also be used and the output is identical.

    This method takes a string as an input and an optional dictionary that map strings to
    integer indices in the input data.

    The string formulas can be specified in several ways. The formula LHS is implicitly
    always ``"y = "`` or ``"y ~ "``, though these can be explicitly provided as well.
    The RHS may contain a set of terms containing the add, multiply, power, and
    call operations much in the same way that the operations would be entered as
    regular python code. Parentheses are used to indicated prececence as well as
    the call operation, and square brackets indicate an indexing operation on the
    inputs. Inputs may be specified as either a string such as ``"x[0]"``,
    ``"inputs[0]"``, or a string that can be mapped to an integer index with the
    optional dictionary passed to the function. Any strings not representing operations
    or inputs as described above are interpreted as follows: if the string can
    be converted into a number, then it is interpreted as a ``ConstantMean`` fixed
    mean function object; otherwise it is assumed to represent a fitting coefficient.
    Note that this means many characters that do not represent operations within this
    mean function language but would not normally be considered as python variables
    will nonetheless be converted into fitting coefficients -- it is up to the user
    to get this right.

    Expressions that are repeated or redundant will not be simplified, so the user should
    take care that the provided expression is sensible as a mean function and will not
    cause problems when fitting.

    Examples: ::

        >>> from mogp_emulator.formula import mean_from_string
        >>> mf1 = mean_from_string("y = a + b*x[0]")
        >>> print(mf1)
        c + c*x[0]
        >>> mf2 = mean_from_string("c*a*b", {"a": 0, "b": 1})
        >>> print(mf2)
        c*x[0]*x[1]

    :param formula: string representing the desired mean function formula
    :type formula: str
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: New subclass of ``MeanBase`` implementing the given formula
    :rtype: subclass of MeanBase (exact type will depend on the formula that is provided)
    """

    assert isinstance(formula, str)

    tokens = _tokenize_string(formula)
    eval_stack = _parse_tokens(tokens)

    mf = _eval_parsed_tokens(eval_stack, inputdict)

    assert issubclass(type(mf), MeanFunction.MeanBase)

    return mf

def _term_to_mean(term, inputdict={}):
    """
    Convert a patsy Term object into a mean function object

    Converts an individual term to a subclass of ``MeanBase`` appropriate for the
    formula (either will be a ``Coefficient`` or a ``MeanProduct`` instance).
    An implicit ``Coefficient`` is included, and then the results from parsing
    each of the factors in the term are multiplied by the coefficient. Variable
    substutions are done using the provided ``inputdict``, mapping strings
    to indices in the inputs, if there is one. Returns the mean function object
    appropriate for this term.

    :param term: patsy ``Term`` object
    :type formula: Term
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: New subclass of ``MeanBase`` implementing the given term
    :rtype: subclass of MeanBase (exact type will depend on the formula that is provided)
    """

    assert not no_patsy, "patsy must be installed to parse formulas using patsy"

    assert isinstance(term, Term)

    # add leading coefficient and multiply by all factors

    mf = MeanFunction.Coefficient()

    for factor in term.factors:
        assert isinstance(factor, EvalFactor)
        mf *= mean_from_string(factor.code, inputdict)

    assert issubclass(type(mf), MeanFunction.MeanBase)

    return mf

def _convert_token(token, inputdict={}):
    """
    Converts an individual token into inputs

    Converts an individual token from the tokenizer to a dependent variable,
    making replacements if the given string has aliases. These include any
    strings found in the provided dictionary and converts ``'inputs'`` to
    ``'x'`` as this is an implicit alias.

    :param token: string token to be converted
    :type str:
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: converted string token if any substitutions were found, otherwise
              the same token is returned
    :rtype: str
    """

    assert isinstance(token, str), "formula input to mean function is not a string"

    if token[:6] == "inputs":
        newtoken = "x"+token[6:]
    elif token in inputdict:
        newtoken = "x[{}]".format(inputdict[token])
    else:
        newtoken = token

    return newtoken

def _is_float(val):
    """
    Checks if a token can be converted into a float

    Returns a boolean indicating if the given token can be converted to a float

    :param val: input to be tested, meant to be a string but can be any type
    :type val: string (though any other type is also acceptable)
    :returns: boolean indicating if input can be converted to a float
    :rtype: bool
    """
    try:
        float(val)
    except ValueError:
        return False
    return True

def _token_to_mean(token, inputdict={}):
    """
    Convert an individual token to a mean function

    Function to convert a single non-operator token to a mean function.
    Makes substitutions found in the provided dictionary (if present)
    to map variable names to indices. Converts numeric tokens to ``ConstantMean``
    objects, tokens that can be converted into inputs to ``LinearMean`` objects,
    and any other strings not found inputdict are assumed to represent
    coefficients and will be converted to ``Coefficient`` objects.

    :param token: string token to be converted
    :type: token: str
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: New subclass of ``MeanBase`` implementing the given token
    :rtype: subclass of MeanBase (exact type will depend on the token)
    """

    assert isinstance(token, str), "formula input to mean function is not a string"

    if _is_float(token):
        return MeanFunction.ConstantMean(float(token))

    token = _convert_token(token, inputdict)

    if not token[0] == "x":
        return MeanFunction.Coefficient()

    if not (token[:2] == "x[" and token[-1] == "]"):
        raise ValueError("bad formula input in mean function")

    try:
        index = int(token[2:-1])
    except ValueError:
        raise ValueError("index in parsed formula is not an integer")

    assert index >= 0, "index in formula parsing must be non-negative"

    return MeanFunction.LinearMean(index)

def _tokenize_string(formula):
    """
    Converts a string formula into a series of tokens for evaluation

    Takes a string and divides it into a list of individual string tokens.
    Special tokens include ``"(", ")", "+", "^", "[", "]", "=", "~"``,
    where these have the expected Python behavior, with ``"="`` and ``"~"``
    separating the LHS and RHS of the formula. All other special characters
    are considered to be non-special in this language and will only be split
    up based on whitespace. The LHS and all white space are removed. Note
    also that ``"call"`` is a protected keyword here, as it is used to
    designate a function call operation.

    :param formula: string formula to be parsed
    :type formula: str
    :returns: list of parsed string tokens
    :rtype: list
    """

    assert isinstance(formula, str)

    token_list = []
    accumulated = ""

    for char in formula:
        if char in ["(", ")", "+", "^", " ", "[", "]", "=", "~"]:
            if not accumulated == "":
                token_list.append(accumulated)
            token_list.append(char)
            accumulated = ""
        elif char == "*":
            if accumulated == "*":
                token_list.append("^")
                accumulated = ""
            elif not accumulated == "":
                token_list.append(accumulated)
                accumulated = "*"
            else:
                accumulated = "*"
        else:
            if accumulated == "*":
                token_list.append(accumulated)
                accumulated = ""
            accumulated += char

    if not accumulated == "":
        token_list.append(accumulated)

    outlist = []

    for item in token_list:
        if not item in [" ", "[", "]"]:
            outlist.append(item)
        elif item == "[":
            outlist.append(outlist.pop()+item)
        elif item == "]":
            if len(outlist) < 2:
                raise SyntaxError("error in using square brackets in formula input")
            outlist.append(outlist.pop(-2)+outlist.pop()+item)

    if outlist[0] == "y":
        outlist.pop(0)
    if outlist[0] in ["=", "~"]:
        outlist.pop(0)

    for item in outlist:
        if (not "[" in item and "]" in item) or (not "]" in item and "[" in item):
            raise SyntaxError("cannot nest operators in square brackets in formula input")
        if item == "call":
            raise SyntaxError("'call' cannot be used as a variable name in formula input")
        if item in ["=", "~"]:
            raise SyntaxError("LHS in formula is not correctly specified")

    return outlist

def _parse_tokens(token_list):
    """
    Parses a list of tokens, converting into an RPN sequence of operations

    Takes a list of string tokens and returns a list of string tokens only
    containing binary operators (``"+", "*", "^", "call"``) and objects on
    which to operate. Order of operations is denoted by Reverse Polish
    Notation, where operations take place on a stack. Uses an adapted
    shunting-yard algorithm.

    :param token_list: list of tokens to be parsed into a sequence of operations
    :type token_list: list
    :returns: list of operators and objects on which to operate
    :rtype: list
    """

    assert isinstance(token_list, list), "input must be a list of strings"

    prev_op = True

    operator_stack = []
    output_list = []

    precedence = {"+": 2, "*": 3, "^": 4}
    l_assoc = {"+": True, "*": True, "^": False}

    for token in token_list:
        assert isinstance(token, str), "input must be a list of strings"
        if token in ["=", "~"]:
            raise SyntaxError("LHS in formula is not correctly specified")
        if not token in ["(", ")", "+", "*", "^"]:
            output_list.append(token)
            prev_op = False
        if token == "(" and not prev_op:
            # function call, put call on the stack
            operator_stack.append("call")
            prev_op = True
        if token in ["+", "*", "^"]:
            while (len(operator_stack) >= 1 and
                   (not operator_stack[-1] == "(") and
                   (operator_stack[-1] == "call" or
                    precedence[operator_stack[-1]] > precedence[token] or
                    (precedence[operator_stack[-1]] == precedence[token] and l_assoc[token]))):
                    output_list.append(operator_stack.pop())
            operator_stack.append(token)
            prev_op = True
        if token == "(":
            operator_stack.append(token)
            prev_op = True
        if token == ")":
            while not operator_stack[-1] == "(":
                output_list.append(operator_stack.pop())
                if len(operator_stack) == 0:
                    raise SyntaxError("string expression has mismatched parentheses")
            if operator_stack[-1] == "(":
                operator_stack.pop()
            prev_op = False

    while len(operator_stack) > 0:
        operator = operator_stack.pop()
        if operator == ")" or operator == "(":
            raise SyntaxError("string expression has mismatched parentheses")
        output_list.append(operator)

    return output_list

def _eval_parsed_tokens(token_list, inputdict={}):
    """
    Evaluate parsed tokens into a mean function

    Evaluates the parsed token list, returning the mean function represented by these
    operations. Input must be a list of strings that come from the ``parse_tokens``
    function, though it will accept any list of strings for processing.

    The function evaluates the sequence of operators and objects on which the operations
    are performed by placing them on a stack and evaluating assuming Reverse Polish
    Notation. The objects are converted to mean function objects when they are placed on
    the stack, and if the stack ever does not contain a valid sequence of operations,
    a ``SyntaxError``is raised. Additionally, if at the end of all operations there is not
    a single mean function object on the stack, a ``SyntaxError`` is raised.

    :param token_list: list of tokens to be parsed into a sequence of operations
    :type token_list: list
    :param inputdict: dictionary used to map variables to input indices. Maps
                      strings to integer indices (must be non-negative). Optional,
                      default is ``{}``.
    :type inputdict: dict
    :returns: New subclass of ``MeanBase`` implementing the given set of operations
    :rtype: subclass of MeanBase (exact type will depend on the sequence)
    """

    assert isinstance(token_list, list), "input must be a list of strings"

    op_list = ["+", "*", "^", "call"]

    stack = []

    for token in token_list:
        assert isinstance(token, str), "tokens must be strings"
        if token in ["=", "~"]:
            raise SyntaxError("LHS in formula is not correctly specified")
        if token not in op_list:
            if token == "I":
                mf = "I"
            else:
                mf = _token_to_mean(token, inputdict)
            stack.append(mf)
        else:
            if len(stack) < 2:
                raise SyntaxError("string expression is not a valid mathematical expression")

            op_2 = stack.pop()
            if op_2 == "I":
                raise SyntaxError("identity operator can only be called as a function")
            assert issubclass(type(op_2), MeanFunction.MeanBase)
            op_1 = stack.pop()
            if token == "call":
                assert op_1 == "I" or issubclass(type(op_1), MeanFunction.MeanBase)
            else:
                if op_1 == "I":
                    raise SyntaxError("identity operator can only be called as a function")
                assert issubclass(type(op_1), MeanFunction.MeanBase)

            if token == "+":
                stack.append(op_1.__add__(op_2))
            elif token == "*":
                stack.append(op_1.__mul__(op_2))
            elif token == "^":
                stack.append(op_1.__pow__(op_2))
            elif token == "call":
                if op_1 == "I":
                    stack.append(op_2)
                else:
                    stack.append(op_1.__call__(op_2))
            else:
                raise SyntaxError("string expression is not a valid mathematical expression")

    if not len(stack) == 1:
        raise SyntaxError("string expression is not a valid mathematical expression")

    return stack[0]




import numpy as np
import mogp_emulator
from projectile import simulator, print_results

# additional GP examples using the projectile demo

# define some common variables

n_samples = 20
n_preds = 10

# Experimental design -- requires a list of parameter bounds if you would like to use
# uniform distributions. If you want to use different distributions, you
# can use any of the standard distributions available in scipy to create
# the appropriate ppf function (the inverse of the cumulative distribution).
# Internally, the code creates the design on the unit hypercube and then uses
# the distribution to map from [0,1] to the real parameter space.

ed = mogp_emulator.LatinHypercubeDesign([(-5., 1.), (0., 1000.)])

# sample space

inputs = ed.sample(n_samples)

# run simulation

targets = np.array([simulator(p) for p in inputs])

###################################################################################

# First example -- fit GP using MLE and Squared Exponential Kernel and predict

print("Example 1: Basic GP")

# create GP and then fit using MLE

gp = mogp_emulator.GaussianProcess(inputs, targets)

gp = mogp_emulator.fit_GP_MAP(gp)

# create 20 target points to predict

predict_points = ed.sample(n_preds)

means, variances, derivs = gp.predict(predict_points)

print_results(predict_points, means)

###################################################################################

# Second Example: How to change the kernel, use a fixed nugget, and create directly using fitting function

print("Example 2: Matern Kernel")

# you can simply pass the args to GP to the fitting function

gp_matern = mogp_emulator.fit_GP_MAP(inputs, targets, kernel='Matern52', nugget=1.e-8)

# return type from predict method is an object with mean, unc, etc defined as attributes

pred_res = gp_matern.predict(predict_points)

print_results(predict_points, pred_res.mean)

###################################################################################

# Third Example: Specify a mean function and set priors to Fit Hyperparameters via MAP

print("Example 3: Mean Function and MAP fitting")

# This example uses a linear mean function and sets priors on the hyperparameters

# Linear mean has 3 hyperparameters (intercept and 2 slopes, one for each input)
# Kernel has 3 hyperparameters (2 correlation lengths, 1 covariance scale)
# Nugget is the final hyperparameter (7 in total)

# Use a normal prior on all mean function values (requires mean, std)
# Use a normal prior on correlation lengths (which are on a log scale, so becomes a lognormal
# distribution once raw values on log scale are converted to linear scale)
# Inverse Gamma distribution on covariance (favors large values)
# Gamma distribution on nugget (favors negative values)

priors = [mogp_emulator.Priors.NormalPrior(0., 10),
          mogp_emulator.Priors.NormalPrior(0., 10.),
          mogp_emulator.Priors.NormalPrior(0., 10.),
          mogp_emulator.Priors.NormalPrior(0., 1.),
          mogp_emulator.Priors.NormalPrior(-10., 1.),
          mogp_emulator.Priors.InvGammaPrior(1., 1.),
          mogp_emulator.Priors.GammaPrior(1., 1.)]

# create GP, passing list of priors and a string representing the mean function
# tell it to estimate the nugget as well

gp_map = mogp_emulator.GaussianProcess(inputs, targets, mean="x[0]+x[1]", priors=priors, nugget="fit")

# fit hyperparameters

gp_map = mogp_emulator.fit_GP_MAP(gp_map)

# gp can be called directly if only the means are desired

pred_means = gp_map(predict_points)

print_results(predict_points, pred_means)


import numpy as np
from numpy.linalg import LinAlgError
from inspect import signature
import warnings

def MH_proposal(current_params, step_sizes):
    """
    Propose an MCMC step using a Metropolis-Hastings method
    
    Proposes the next point in an MCMC sampler using the Metropolis-Hastings method.
    Inputs are the current point and a covariance matrix. The next point is drawn from
    a multivariate normal distribution centered around the current point. The covariance
    matrix must be a 2D ``n`` by ``n`` array, where ``n`` is the number of parameters, and
    must be positive definite. Returns a 1D array of length ``n`` holding the new proposed
    parameter values.
    
    :param current_params: Current value of the parameters. Must be a 1D array.
    :type current_params: ndarray
    :param step_sizes: Covariance matrix from which steps are drawn. Must be a 2D array
                       with both dimensions the same length as ``current_params``, and
                       must be positive definite.
    :type step_sizes: ndarray
    :returns: New value of parameters, a 1D array with the same length as ``current_params``
    :rtype: ndarray
    """
    
    current_params = np.array(current_params)
    step_sizes = np.array(step_sizes)
    
    assert len(current_params.shape) == 1, "current parameters must be a 1D array"
    assert len(step_sizes.shape) == 2, "step sizes must be a 2D array"
    assert step_sizes.shape[0] == step_sizes.shape[1], "step sizes must be a square array"
    assert len(current_params) == step_sizes.shape[0], "length of current parameters must match length of step sizes"
    assert np.all(np.diag(step_sizes) > 0.), "step sizes must be a positive definite matrix"
    
    return np.random.multivariate_normal(mean=current_params, cov = step_sizes)


def MCMC_step(loglikelihood, current_params, step_sizes, loglike_sign = 1.):
    """
    Method to take a weak prior Metropolis-Hastings MCMC step
    
    Take a Metropolis-Hastings MCMC step with given log-likelihood function and step sizes.
    Method uses a multivariate normal distribution centered around the current parameter
    values from which the next point is drawn, and evaluates the log-likelihood for both
    points. If the next point has a larger log-likelihood, the step is always accepted,
    and if the log-likelihood is less for the proposed step it is accepted with a
    probability based on the difference between the two. If the method encounters an
    error when evalutaing the log-likelihood, the step is rejected. Returns the next point
    and a boolean indicating whether or not the step was accepted.
    
    An optional parameter ``loglike_sign`` can be passed that must be a float with the
    value +/- 1. This is multiplied by the log-likelihood and thus allows methods that
    compute the negative log-likelihood to be used in this routine. If values other than
    +/- 1 are passed, the method will raise an error.
    
    :param loglikelihood: Log-likelihood function to be used in the MCMC step. Must be
                          callable and must accept a single argument, which is the array
                          holding the parameters. If this function computes the negative
                          log-likelihood, pass ``loglike_sign = -1.`` to the function as well.
    :type loglikelihood: function or other callable
    :param current_params: Current value of the parameters. Must be a 1D array.
    :type current_params: ndarray
    :param step_sizes: Covariance matrix from which steps are drawn. Must be a 2D array
                       with both dimensions the same length as ``current_params``, and
                       must be positive definite.
    :type step_sizes: ndarray
    :param loglike_sign: Sign for the log-likelihood function. If the provided
                         ``loglikelihood`` function computes the negative log-likelihood,
                         pass ``-1.`` for this parameter. Optional, default value is ``1.``
    :type loglike_sign: float
    :returns: Proposed next point and whether or not the point is accepted as a tuple.
              The first return item is the next point (as a 1D array with the same length\
              as ``current_params``) and the second item is a boolean indicating whether
              or not the step was accepted.
    :rtype: tuple containing a ndarray and a bool
    """
    
    assert callable(loglikelihood), "loglikelihood must be a callable function"
    assert len(signature(loglikelihood).parameters) == 1
    assert loglike_sign == 1. or loglike_sign == -1., "loglikelihood sign must be +/- 1"
    
    next_point = MH_proposal(current_params, step_sizes)

    try:
        H = loglike_sign*(-loglikelihood(current_params) + loglikelihood(next_point))
    except (FloatingPointError, AssertionError, LinAlgError):
        H = np.nan

    if H >= np.log(np.random.random()) and np.isfinite(H):
        accept = True
    else:
        accept = False

    return next_point, accept

def sample_MCMC(loglikelihood, start, step_sizes, n_samples = 1000, thin = 0, loglike_sign = 1.):
    """
    Draw MCMC samples for a given log-likelihood function with weak priors
    
    Compute an MCMC chain for a given log-likelihood function with weak priors. Function
    requires the log-likelihood function, the starting point for the MCMC chain, and an
    array describing the step sizes. Optional parameters are the number of steps to take,
    how the thin the samples, and a sign for the log-likelihood function.
    
    The log-likelihood function must be a function or other callable that accepts a single
    argument, which is a 1D array holding the current parameter values. The starting point
    must be a 1D array holding the starting parameter values, which must match the length of
    the input to the log-likelihood function. The step sizes must be a 2D array with each
    dimension having the same length as the number of parameters, and must be positive
    definite. The step size array is used as the covariance matrix for a multivariate
    normal distribution, from which steps are drawn.
    
    Optional parameters are the number of steps to be taken (must be a positive integer).
    Note that if the chain is thinned, the number of points in the final MCMC chain will
    differ from the number of steps taken.
    
    Thinning may be specified with a non-negative integer. If a positive integer is
    given, the chain will be thinned by only keeping every ``thin`` steps. Note that
    ``thin = 1`` means that the chain will not be thinned. If ``thin = 0`` is given
    (the default value), the chain will automatically be thinned by computing the
    autocorrelation of the chain for each parameter separately and estimating the value
    needed to eliminate correlations in the chain. If the autothinning method fails
    (usually occurrs if the posterior is multimodal), the chain will not be thinned
    and a warning will be given. More details on the autothinning procedure are
    described in the corresponding function.
    
    An optional parameter ``loglike_sign`` can be passed that must be a float with the
    value +/- 1. This is multiplied by the log-likelihood and thus allows methods that
    compute the negative log-likelihood to be used in this routine.
    
    Returns the final thinned MCMC chain (a 2D array, where the first dimension indicates
    the different samples and the second dimension indicates the different parameters),
    an array holding all rejected steps (also a 2D array like the MCMC chain, useful for
    diagnosing problems with convergence), the fraction of steps that are accepted (also
    useful for diagnosing problems with convergence), and the first lag autocorrelation of
    the thinned MCMC chain.
    
    :param loglikelihood: Log-likelihood function to be used in the MCMC step. Must be
                          callable and must accept a single argument, which is the array
                          holding the parameters. If this function computes the negative
                          log-likelihood, pass ``loglike_sign = -1.`` to the function as well.
    :type loglikelihood: function or other callable
    :param start: Starting value of the parameters. Must be a 1D array.
    :type start: ndarray
    :param step_sizes: Covariance matrix from which steps are drawn. Must be a 2D array
                       with both dimensions the same length as ``current_params``, and
                       must be positive definite.
    :type step_sizes: ndarray
    :param n_samples: Number of steps to be taken. Must be a positive integer. Optional,
                      default value is 1000. Note that if the chain is thinned, the
                      final MCMC chain will be shorter than this.
    :type n_samples: int
    :param thin: Integer describing how to thin the MCMC chain. A positive integer
                 indicates manual thinning by keeping every ``thin`` steps (note
                 that ``thin = 1`` means the chain will not be thinned). A value
                 of ``0`` will attempt to autothin the chain.
    :type thin: int
    :param loglike_sign: Sign for the log-likelihood function. If the provided
                         ``loglikelihood`` function computes the negative log-likelihood,
                         pass ``-1.`` for this parameter. Optional, default value is ``1.``
    :type loglike_sign: float
    :returns: MCMC chain (2D array), array of rejected points (2D array), acceptance rate
              (float), and first lag autocorrelation of the thinned MCMC chain (float)
    :rtype: tuple containing (ndarray, ndarray, float, float)
    """

    n_samples = int(n_samples)
    thin = int(thin)

    assert n_samples > 0, "number of samples must be a positive integer"
    assert thin >= 0, "thin must be a non-negative integer"
    
    start = np.array(start)
    
    assert start.ndim == 1, "starting point must be a 1d array"
    n_params = len(start)

    samples = np.zeros((n_samples, n_params))
    samples[0] = start
    rejected = []

    for i in range(n_samples - 1):
        next_point, accept = MCMC_step(loglikelihood, samples[i], step_sizes, loglike_sign)
        if accept:
            samples[i+1] = np.copy(next_point)
        else:
            samples[i+1] = np.copy(samples[i])
            rejected.append(np.array(next_point))
        
    acceptance = float(n_samples - len(rejected))/float(n_samples)

    if thin == 0:
        thin_freq = autothin_samples(samples)
    else:
        thin_freq = thin
    
    thinned = samples[::thin_freq]

    first_lag = np.zeros(n_params)

    if n_samples > 1:
        for i in range(n_params):
            autocorr = np.correlate(thinned[:,i]-np.mean(thinned[:,i]), thinned[:,i]-np.mean(thinned[:,i]), mode="full")
            if np.max(autocorr) > 0.:
                first_lag[i] = autocorr[np.argmax(autocorr)+1]/np.max(autocorr)

    return thinned, np.array(rejected), acceptance, first_lag

def autothin_samples(signal):
    """
    Automatically estimate thinning needed to obtain uncorrelated samples
    
    This function attempts to estimate the thinning needed to obtain uncorrelated samples in an
    MCMC chain. For each separate parameter, the function computes the autocorrelation and
    estimates the lag needed to obtain uncorrelated samples. This is done by recognizing that
    the standard deviation of the autocorrelation of a random signal scales inversely with the
    square root of the number of samples, so when the autocorrelation drops below three times
    this, we use this as a guess to when the signal is uncorrelated. The maximum lag across all
    parameters is returned. If the chain contains fewer than 10 points, or the autocorrelation
    never drops below the target value, the method gives a warning and
    returns 1.
    
    :param signal: MCMC chain to be thinned. Must be a 1D or 2D array. If 2D, the first dimension
                   indicates the MCMC samples, and the second dimension indicates the different
                   parameter values. If 1D, the array is assumed to hold samples of a single
                   parameter, and the array is reshaped to have a singleton second dimension.
    :type signal: ndarray
    :returns: Estimated stride needed to thin the samples to obtain uncorrelated samples. If
              the method does not succeed, gives a warning and returns ``1``.
    :rtype: int
    """
    
    signal = np.array(signal)
    if signal.ndim == 1:
        signal = np.reshape(signal, (len(signal), 1))

    assert signal.ndim == 2, "input signal to be thinned must be a 1d or 2d array"
    
    n_samples, n_params = signal.shape
    
    maxthin = 0
    
    if n_samples >= 10:
        for i in range(n_params):
            autocorr = np.correlate(signal[:,i]-np.mean(signal[:,i]), signal[:,i]-np.mean(signal[:,i]), mode="full")
            start = np.argmax(autocorr)
            if np.max(autocorr) > 0.:
                autocorr = autocorr/np.max(autocorr)
            for j in range(1, len(autocorr) - start):
                if np.abs(autocorr[start + j]) < 3./np.sqrt(float(n_samples)):
                    if j > maxthin:
                        maxthin = j
                    break
    
    if maxthin == 0:
        warnings.warn("automatic thinning failed, posterior distribution may be multimodal")
        maxthin = 1
        
    return maxthin
    


'''
This benchmark examines the performance of emulators fit in parallel to multiple targets. The
benchmark uses a set of tsunami simulations, where the inputs are 14 values of the seafloor
displacement resulting from an earthquake, and the outputs are tsunami wave heights at
different spatial locations. This benchmark fits 8, 16, 32, and 64 output points using
1, 2, 4, and 8 processes, and records the time required per emulator to perform the fitting.
The actual performance will depend on the specific machine and the number of cores available.
Once the number of processes exceeds the number of cores on the machine, the fitting time will
increase, so the results will depend on the exact setup used. For reference, tests on a quad core
MacBook Pro found that the fitting took roughly 1 second per emulator on a single core, with the time
per emulator dropping by about a factor of 2 when 4 processes were used.
'''

import numpy as np
from mogp_emulator import MultiOutputGP, fit_GP_MAP
from time import time
try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

def load_tsunami_data(n_emulators):
    "loads tsunami data from npz archive"

    assert(n_emulators > 0)

    f = np.load('tsunamidata.npz')

    assert('inputs' in f.files)
    assert('targets' in f.files)
    assert(f['inputs'].shape[0] == f['targets'].shape[1])

    return f['inputs'], f['targets'][:n_emulators]

def fit_emulators(n_emulators, processes = None):
    "load data and fit emulators for tsunami data, returning the time required to fit the emulators"

    inputs, targets = load_tsunami_data(n_emulators)

    gp = MultiOutputGP(inputs, targets)

    start_time = time()
    gp = fit_GP_MAP(gp, processes = processes)
    finish_time = time()

    return finish_time - start_time

def make_scaling_plot(n_emulators_list, process_list, execution_times):
    "create scaling plot showing time for fitting tsunami emulators"

    fig = plt.figure(figsize=(4,3))
    for n_emulators, exec_times in zip(n_emulators_list, execution_times):
        plt.plot(process_list, exec_times/float(n_emulators), '-o', label = str(n_emulators)+" emulators")
    plt.legend()
    plt.xlabel('Number of processes')
    plt.ylabel('Execution time per emulator (s)')
    plt.title("Benchmark on tsunami simulation data")
    plt.savefig('tsunami_scaling.png', bbox_inches='tight')

def run_all_models(n_emulators_list, process_list):
    "Run all sets of emulators with varying number of processes"

    execution_times = []

    for n_emulators in n_emulators_list:
        emulator_execution_times = []
        for processes in process_list:
            emulator_execution_times.append(fit_emulators(n_emulators, processes = processes))
        execution_times.append(np.array(emulator_execution_times))

    print("\n")
    print("Num. Emulators    Num. Processors    Execution Time (s)   Execution Time per Emulator (s)")
    for n_emulators, exec_times in zip(n_emulators_list, execution_times):
        for process, exec_time in zip(process_list, exec_times):
            print("{:18}{:19}{:21}{}".format(str(n_emulators), str(process), str(exec_time), str(exec_time/float(n_emulators))))

    if makeplots:
        make_scaling_plot(n_emulators_list, process_list, execution_times)

if __name__ == "__main__":
    n_emulators_list = [8, 16, 32, 64]
    process_list = [1, 2, 4, 8]
    run_all_models(n_emulators_list, process_list)



'''
This benchmark performs convergence tests on multiple realizations of the 2D Branin function.
Details of the 2D Branin function can be found at https://www.sfu.ca/~ssurjano/branin.html.
This particular version uses 8 realizations of the Branin function, each with a different
set of parameters. The code samples these 8 realizations simultaneously using a spacefilling
Latin Hypercube experimental design with a varying number of target points, and then tests
the convergence of the resulting emulators. As the number of targe points increases, the
prediction error and prediction variance should decrease.

(Note however that eventually, the predictions worsen once the number of target points becomes
large enough that the points become too densely sampled. In this case, the points become
co-linear and the resulting covariance matrix is singular and cannot be inverted. To avoid
this problem, the code iteratively adds additional noise to the covariance function to
stabilize the inversion. However, this noise reduces the accuracy of the predictions. The
values chosen for this benchmark attempt to avoid this, but in some cases this still becomes
a problem due to the inherent smoothness of the squared exponential covariance function.)
'''

import numpy as np
from mogp_emulator import MultiOutputGP, fit_GP_MAP
from mogp_emulator import MonteCarloDesign, LatinHypercubeDesign
from scipy.stats import uniform
try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

def branin_2d(x1, x2, params):
    "2D Branin function, see https://www.sfu.ca/~ssurjano/branin.html for more information"
    a, b, c, r, s, t = params
    return a*(x2 - b*x1**2 + c*x1 - r)**2 + s*(1. - t)*np.cos(x1) + s

def generate_emulator_params(n_emulators):
    "Generate random parameters for use with 2D Branin function, ensuring outputs are similarly valued"

    n_emulators = int(n_emulators)
    assert n_emulators > 0

    x1 = np.linspace(-5., 10.)
    x2 = np.linspace(0., 15.)
    x1m, x2m = np.meshgrid(x1, x2, indexing='ij')

    params = np.zeros((n_emulators, 6))

    i = 0

    while i < n_emulators:

        a = np.random.normal(loc = 1., scale = 0.1)
        b = np.random.normal(loc = 5.1/4./np.pi**2, scale = 0.1)
        c = np.random.normal(loc = 5./np.pi, scale = 0.1)
        r = np.random.normal(loc = 6., scale = 1.)
        s = np.random.normal(loc = 10., scale = 2.)
        t = np.random.normal(loc = 1./8./np.pi, scale = 0.01)

        branin_vals = branin_2d(x1m, x2m, (a, b, c, r, s, t))

        if np.all(branin_vals >= 0.) and np.all(branin_vals <= 350.):
            params[i] = np.array([a, b, c, r, s, t])
            i += 1

    return params

def generate_input_data(n_simulations, method = "random"):
    "Generate random points x1 and x2 for evaluating the multivalued 2D Branin function"

    n_simulations = int(n_simulations)
    assert(n_simulations > 0)
    assert method == "random" or method == "lhd"

    if method == "random":
        ed = MonteCarloDesign([uniform(loc = -5., scale = 15.).ppf, uniform(loc = 0., scale = 15.).ppf])
    elif method == "lhd":
        ed = LatinHypercubeDesign([uniform(loc = -5., scale = 15.).ppf, uniform(loc = 0., scale = 15.).ppf])
    inputs = ed.sample(n_simulations)
    return inputs

def generate_target_data(inputs, emulator_params):
    "Generate target data for multivalued emulator benchmark"

    emulator_params = np.array(emulator_params)
    assert emulator_params.shape[1] == 6
    n_emulators = emulator_params.shape[0]

    inputs = np.array(inputs)
    assert len(inputs.shape) == 2
    assert inputs.shape[1] == 2
    n_simulations = inputs.shape[0]

    targets = np.zeros((n_emulators, n_simulations))

    for i in range(n_emulators):
        targets[i] = branin_2d(inputs[:,0], inputs[:,1], emulator_params[i])

    return targets

def generate_training_data(n_emulators, n_simulations):
    "Generate n_simulations input data and evaluate using n_emulators different parameter values"

    emulator_params = generate_emulator_params(n_emulators)
    inputs = generate_input_data(n_simulations, method = "lhd")
    targets = generate_target_data(inputs, emulator_params)

    return inputs, targets, emulator_params

def generate_test_data(n_testing, emulator_params):
    "Generate n_testing points for testing the accuracy of an emulator"

    testing = generate_input_data(n_testing, method = "random")
    test_targets = generate_target_data(testing, emulator_params)

    return testing, test_targets

def run_model(n_emulators, n_simulations, n_testing, processes = None):
    "Generate training data, fit emulators, and test model accuracy on random points, returning RMSE"

    inputs, targets, emulator_params = generate_training_data(n_emulators, n_simulations)

    gp = MultiOutputGP(inputs, targets)
    gp = fit_GP_MAP(gp, processes = processes)

    norm_const = np.mean(targets)

    testing, test_targets = generate_test_data(n_testing, emulator_params)

    test_vals, unc, deriv = gp.predict(testing, deriv = False, unc = True, processes = processes)

    return (np.sqrt(np.sum((test_vals - test_targets)**2)/float(n_emulators)/float(n_testing))/norm_const,
            np.sqrt(np.sum(unc**2)/float(n_emulators)/float(n_testing))/norm_const**2)

def plot_model_errors(simulation_list, error, unc, n_testing, n_emulators):
    "Makes plot showing accuracy of emulator as a function of n_simulations"

    plt.figure(figsize=(4,3))
    plt.semilogy(simulation_list, error,'-o')
    plt.xlabel('Number of simulations')
    plt.ylabel('Average prediction RMSE')
    plt.title('Error for '+str(n_testing)+' predictions\nusing '+str(n_emulators)+' 2D Branin functions')
    plt.savefig('branin_2d_error.png',bbox_inches='tight')

    plt.figure(figsize=(4,3))
    plt.semilogy(simulation_list, unc,'-o')
    plt.xlabel('Number of simulations')
    plt.ylabel('Average prediction variance')
    plt.title('Uncertainty for '+str(n_testing)+' predictions\nusing '+str(n_emulators)+' 2D Branin functions')
    plt.savefig('branin_2d_unc.png',bbox_inches='tight')

def run_all_models(n_emulators, n_testing, simulation_list, process_list = [None], n_iter = 10):
    "Runs all models, printing out results and optionally making plots"

    n_simtrials = len(simulation_list)

    errors = np.zeros((n_simtrials, n_iter))
    uncs = np.zeros((n_simtrials, n_iter))

    for processes in process_list:
        for iteration in range(n_iter):
            for sim_index in range(n_simtrials):
                errors[sim_index, iteration], uncs[sim_index, iteration] = run_model(n_emulators, simulation_list[sim_index], n_testing, processes)

    error = np.mean(errors, axis = -1)
    unc = np.mean(uncs, axis = -1)

    print("\n")
    print("Convergence test results:")
    print("Num. simulations   Average prediction RMSE")
    for sim, err in zip(simulation_list, error):
        print('{:19}{}'.format(str(sim), str(err)))

    print("\n")
    print("Num. simulations   Average prediction variance")
    for sim, un in zip(simulation_list, unc):
        print('{:19}{}'.format(str(sim), str(un)))

    if makeplots:
        plot_model_errors(simulation_list, error, unc, n_testing, n_emulators)

if __name__ == '__main__':
    run_all_models(8, 100, [int(x) for x in np.linspace(10., 30., 11)], process_list = [4], n_iter = 10)